from scc.constants import SCButtons

BUTTON_ORDER = (
	SCButtons.A, SCButtons.B, SCButtons.X, SCButtons.Y,
	SCButtons.BACK, SCButtons.C, SCButtons.START,
	SCButtons.LB, SCButtons.RB, SCButtons.LT, SCButtons.RT,
	SCButtons.STICKPRESS, SCButtons.LPAD, SCButtons.RPAD,
	SCButtons.RGRIP, SCButtons.LGRIP
)
